__title__ = 'pybullet_planning'
__description__ = 'a suite of utility functions to facilitate robotic planning related research on the pybullet physics simulation engine.'
__url__ = 'https://github.com/yijiangh/pybullet_planning'
__version__ = '0.6.1'
__author__ = 'Caelan Garrett'
__author_email__ = 'yijiangh@mit.edu' # this is only for the purpose of pypi...
# __author_email__ = 'caelan@csail.mit.edu; '
__license__ = 'MIT license'
__copyright__ = 'Copyright 2019 Caelan Garrett'

__all__ = ['__author__', '__author_email__', '__copyright__', '__description__', '__license__', '__title__', '__url__', '__version__']
